# ParameterPartialOrders.py
# Marcio Gameiro
# 2020-10-08
# MIT LICENSE

import DSGRN
import numpy as np

def hex2partial(hex_code, n_inputs, n_outputs):
    # Computes the partial order corresponding to a
    # given hex code. This can probably be optimized.

    n_bits = n_outputs * (2**n_inputs) # number of bits
    fmt = '0' + str(n_bits) + 'b' # bit code format
    par_bit_code = format(int(hex_code, 16), fmt) # get the parameter bit code

    par_bit_array = list(reversed(np.fromiter(map(int, par_bit_code), dtype=int))) # array of bit codes
    # transform into matrix using column-major order (Fortran-like index ordering)
    par_bit_matrix = np.reshape(par_bit_array, (n_outputs, -1), order='F') # matrix of bit codes

    partial_order = [] # partial order list
    total_order_vals = list(range(2**n_inputs)) # values used in total order

    for row_indx, row in enumerate(par_bit_matrix): # for each row
        for col_indx, bit_val in enumerate(row): # for each column
            if (bit_val == 0) and (col_indx in total_order_vals): # if bit==0 and have not yet added col_indx
                total_order_vals.remove(col_indx) # remove from list of values not yet added
                partial_order.append(col_indx) # add value to the list
        partial_order.append(-(n_outputs - row_indx)) # add negative value for threshold

    partial_order.extend(total_order_vals) # add remaning values to list

    return partial_order

def factor_graph_inequalities(parameter_graph, par_index, node):
    # Given a parameter index and a network node returns the
    # parameter inequalities (as a partial order tuple) of the
    # factor graph corresponding to this node.
    parameter = parameter_graph.parameter(par_index)
    # Get inputs and outputs to node
    inputs = parameter_graph.network().inputs(node)
    outputs = parameter_graph.network().outputs(node)
    # Get number of inputs and outputs
    n_inputs = len(inputs)
    n_outputs = len(outputs)
    # Get factor graph hex code
    hex_code = parameter.logic()[node].hex()
    # Compute partial order corresponding to hex code
    partial_order = hex2partial(hex_code, n_inputs, n_outputs)

    def threshold_str(source_node, original_edge_index):
        # Given a source node and an edge index returns a
        # string for the threshold corresponding to the
        # out edge from source to outputs[edge_index] after
        # applying the order parameter permutation to edge_index.
        #
        # Get source node name
        source_name = parameter_graph.network().name(source_node)
        # Apply order parameter permutation to original_edge_index
        edge_index = parameter.order()[source_node](original_edge_index)
        # Get target node and name
        target = outputs[edge_index]
        target_name = parameter_graph.network().name(target)
        thres_str = 'T[' + source_name + '->' + target_name + ']'
        return thres_str

    # Make string for a partial order term
    def partial_order_term_str(p, source_node):
        if p < 0: # Return string for threshold
            # Get original edge index (out edge order)
            # This is the order before permutation
            original_edge_index = n_outputs + p
            return threshold_str(source_node, original_edge_index)
        # Return string for polynomial
        return 'p' + str(p)

    # Get node name
    node_name = parameter_graph.network().name(node)
    # Make inequalities string
    inequalities = node_name + ' : '
    first = True
    for p in partial_order:
        # Get partial order term string
        p_str = partial_order_term_str(p, node)
        if first:
            inequalities += '(' + p_str
            first = False
        else:
            inequalities += ', ' + p_str
    inequalities += ')'

    return inequalities

def parameter_partial_orders(parameter_graph, par_index):
    D = parameter_graph.dimension()
    # Get parameter inequalities
    parameter_ineqs = ''
    for d in range(D):
        # Get inequalities for each factor graph
        factor_ineqs = factor_graph_inequalities(parameter_graph, par_index, d)
        parameter_ineqs += factor_ineqs + '\n'
    # Remove last '\n' and return
    return parameter_ineqs.strip()

def parameter_output_thresholds(parameter, node):
    # Get outputs to node
    outputs = parameter.network().outputs(node)
    # Get number of outputs
    n_outputs = len(outputs)
    # Get output thresholds
    output_thresholds = ''
    for j in range(n_outputs):
        out_thres = parameter.output_threshold(j, node)
        output_thresholds += out_thres + '\n'
    # Remove last '\n' and return
    return output_thresholds.strip()

def parameter_input_polynomials(parameter, node):
    # Get inputs to node
    inputs = parameter.network().inputs(node)
    # Get number of inputs
    n_inputs = len(inputs)
    # Number of input combinations
    n_in_combs = 2**n_inputs
    # Get input polynomials
    input_polynomials = ''
    for i in range(n_in_combs):
        in_poly = parameter.input_polynomial(i, node)
        input_polynomials += in_poly + '\n'
    # Remove last '\n' and return
    return input_polynomials.strip()



import sqlite3
import graphviz
from DSGRN._dsgrn import *
from functools import reduce

from DSGRN.Query.Logging import LogToSTDOUT

class Database:
  def __init__(self, database_name):
    """
    Initialize a DSGRN database object
    """
    self.dbname = database_name
    self.conn = sqlite3.connect(database_name)
    self.cursor = self.conn.cursor()
    # Load network spec from database
    sqlexpression = "select Specification from Network"
    self.cursor.execute(sqlexpression)
    network_spec = self.cursor.fetchone()[0]
    # construct network
    self.network = Network(network_spec)
    self.parametergraph = ParameterGraph(self.network)
    # D is the number of network nodes
    self.D = self.parametergraph.dimension()
    self.names = [ self.network.name(i) for i in range(0, self.D)]
    # DSGRN uses an indexing scheme to refer to parameters. It is based on a mixed-radix number scheme
    # where the place value of each digit varies according to the number of logic parameters for each node
    # and the number of order parameter for each node. Specifically, the ordering of the digits is (from least
    # significant) the sizes of each factor graph, followed by the number of permutations of the out-edges for
    # each node. We call these "bases" (as in number base) and we compute the place value for each digit.
    self.indexing_place_bases = [self.parametergraph.logicsize(i) for i in range(0,self.D)] + [self.parametergraph.ordersize(i) for i in range(0,self.D)]
    self.indexing_place_values = reduce ( lambda x, y : x + [x[-1]*y], self.indexing_place_bases[:-1], [1])

  def execute(self, expression, parameters = None):
    """
    Perform an SQL query.
    Returns a "cursor" object (see python sqlite3 API for details)
    """
    return self.cursor.execute(expression, parameters)

  def __call__(self, pi):
    c = self.conn.cursor()
    sqlexpression = "select MorseGraphIndex from Signatures where ParameterIndex = ?"
    c.execute(sqlexpression,(pi,))
    mgi = c.fetchone()[0]
    return mgi

  def __del__(self):
    """
    Commit and close upon destruction
    """
    self.conn.commit()
    self.conn.close()

  def _repr_svg_(self):
    return graphviz.Source(self.network.graphviz())._repr_svg_()

  def DrawMorseGraph(self, morsegraphindex):
    """
    Return an object which renders to a graphviz representation in Jupyter
    """
    c = self.conn.cursor()
    sqlexpression = "select Graphviz from MorseGraphViz where MorseGraphIndex = ?"
    c.execute(sqlexpression,(morsegraphindex,))
    gv = c.fetchone()[0]
    return graphviz.Source(gv)




class BistableQuery:
  """
  BistableQuery
    Query to check if morse graph index corresponds to a Morse graph with precisely two minimal elements
  """
  def __init__ (self, database ):
    self.database = database
    if not hasattr(database, 'BistableQuery'):
      c = database.conn.cursor()
      sqlexpression = "select MorseGraphIndex from (select MorseGraphIndex, count(*) as StableCount from (select MorseGraphIndex,Vertex from MorseGraphVertices except select MorseGraphIndex,Source from MorseGraphEdges) group by MorseGraphIndex) where StableCount=2;"
      database.BistableQuery = frozenset([ row[0] for row in c.execute(sqlexpression) ])

  def matches(self):
    """
    Return entire set of matches
    """
    return self.database.BistableQuery

  def __call__ (self, morsegraphindex ):
    """ 
    Test if a single mgi is in the set of matches
    """
    return morsegraphindex in self.database.BistableQuery



# DoubleFixedPointQuery.py
# MIT LICENSE 2016
# Shaun Harker

from DSGRN.Query.Logging import LogToSTDOUT
from DSGRN.Query.FixedPointTables import *

class DoubleFixedPointQuery:
  def __init__ (self, database, bounds1, bounds2):
    self.database = database
    MatchQuery(bounds1,"Matches1",database);
    MatchQuery(bounds2,"Matches2",database);
    c = database.conn.cursor()
    c.execute('create temp table Left as select MorseGraphIndex ' +
              'from (select * from Matches1 except select * from Matches2) group by MorseGraphIndex;')
    c.execute('create temp table Middle as select MorseGraphIndex, count(Vertex) as middle ' +
              'from (select * from Matches1 intersect select * from Matches2) group by MorseGraphIndex;')
    c.execute('create temp table Right as select MorseGraphIndex ' +
              'from (select * from Matches2 except select * from Matches1) group by MorseGraphIndex;')
    c.execute('create temp table Matches as ' +
              'select * from (select * from Left intersect select * from Right) ' +
              'union ' +
              'select * from (select MorseGraphIndex from Middle ' +
                             'intersect ' +
                             'select * from (select * from Left union select * from Right)) ' +
              'union ' +
              'select MorseGraphIndex from Middle where middle >= 2;')

    # Final query and print results
    self.set_of_matches = set([ row[0] for row in c.execute('select MorseGraphIndex from Matches;')])
    #result = [ row[0] for row in c.execute('select distinct ParameterIndex from Matches natural join Signatures order by ParameterIndex;')]
    c.execute('drop table Matches1');
    c.execute('drop table Matches2');
    c.execute('drop table Matches');
    c.execute('drop table Left');
    c.execute('drop table Middle');
    c.execute('drop table Right');

  def matches(self):
    """
    Return entire set of matches
    """
    return self.set_of_matches

  def __call__ (self, morsegraphindex ):
    """ 
    Test if a single mgi is in the set of matches
    """
    return morsegraphindex in self.set_of_matches



# MultistableQuery.py
# MIT LICENSE 2016
# Shaun Harker

class MultistableQuery:
  """
  MultistableQuery
    Query to check if morse graph index corresponds to a Morse graph with more than one minimal element
  """
  def __init__ (self, database ):
    self.database = database
    if not hasattr(database, 'MultistableQuery'):
      c = database.conn.cursor()
      sqlexpression = "select MorseGraphIndex from (select MorseGraphIndex, count(*) as StableCount from (select MorseGraphIndex,Vertex from MorseGraphVertices except select MorseGraphIndex,Source from MorseGraphEdges) group by MorseGraphIndex) where StableCount>1;"
      database.MultistableQuery = frozenset([ row[0] for row in c.execute(sqlexpression) ])

  def matches(self):
    """
    Return entire set of matches
    """
    return self.database.MultistableQuery

  def __call__ (self, morsegraphindex ):
    """ 
    Test if a single mgi is in the set of matches
    """
    return morsegraphindex in self.database.MultistableQuery



### BlowUpGraph.py
### Shaun Harker
### 2018-03-25
### MIT LICENSE
#
# Marcio Gameiro
# 2021-03-26

from pychomp2 import *
import datetime
import operator
import functools
from itertools import product
from math import log2
from DSGRN._dsgrn import *

class BlowupGraph:

    def complex(self):
        return self.dc
        
    def diagram(self):
        return self.digraph.adjacencies

    def closedrightfringe(self,i):
        return any(self.cc.rightfringe(k) for k in self.cc.star({i}))

    def blowupfringe(self,i):
        """
        determine if cell i is in the fringe
        """
        return all(self.closedrightfringe(k) for k in self.oc.simplex(self.dc.dual(i)))

    def NormalVariables(self,s):  # i.e. inessential directions
        """
        return the set of normal variables for a cubical cell s
        """
        shape = self.cc.cell_shape(s)
        return [ d for d in range(0,self.D) if shape & (1 << d) == 0]

    def TangentVariables(self,s):  # i.e. inessential directions
        """
        return the set of normal variables for a cubical cell s
        """
        shape = self.cc.cell_shape(s)
        return [ d for d in range(0,self.D) if shape & (1 << d) != 0]

    def RelativePositionVector(self,s,t):
        """
        Return the relative position vector of two cubical cells;
        this is equal to the difference of combinatorial position of two cells,
        regarding vertices to be at (even,even,...) integer coordinates and
        higher dimensional cells to have odd coordinates for their dimensions with
        extent
        """
        return [ y-x for (x,y) in zip(self.cc.barycenter(s),self.cc.barycenter(t))]

    def absorbing(self, cctopcell, d, direction):
        """
        Return True if the codim 1 face of cctopcell 
          collapsed in dimension d in direction "direction"
          is absorbing
        """
        # The cc topcell indexing is not compatible with labelling indexing
        # due to the extra wrapped layer, so we compute the index idx
        coords = self.cc.coordinates(cctopcell)
        # "Wrap layer" boxes have all walls absorbing.
        if any(coords[i] == self.limits[i] for i in range(0,self.D)):
            return True
        idx = sum (  c * self.pv[i] for (i,c) in enumerate(coords) )
        labelbit = 1 << (d + (self.D if direction == -1 else 0))
        return self.labelling[idx] & labelbit != 0

    def flowdirection(self, cctopcell, s, t):
        """
        Given a cubical domain, and a dual-order wall (s,t) (i.e. s is a face of t in cc) 
        return 1 if a proof of transversality from s->t is known valid in dom
        return -1 if a proof of transversality from t->s is known valid in dom
        return 0 otherwise
        """
        #print("flowdirection" + str((cctopcell, s, t)))
        sshape = self.cc.cell_shape(s)
        scoords = self.cc.coordinates(s)
        tshape = self.cc.cell_shape(t)
        tcoords = self.cc.coordinates(t)
        xorshape = sshape ^ tshape # 1's in positions where s is 0 and t is 1.
        absorbing = False
        repelling = False
        #print("Line 40" + str((sshape, scoords, tshape, tcoords, xorshape)))
        for d in range(0,self.D):
            if xorshape & (1 << d):
                #print("  collapsed dimension " + str(d))
                direction = -1 if scoords[d] > tcoords[d] else 1  # -1 left, 1 right
                #print("  direction " + str(direction))
                #print("  absorbing -> " + str(self.absorbing(cctopcell, d, direction)))
                if self.absorbing(cctopcell, d, direction):
                    absorbing = True
                else:
                    repelling = True
        if absorbing and not repelling:
            return -1
        elif repelling and not absorbing:
            return 1
        else:
            return 0

    def RookField(self,t):
        """
        Given a cubical top-cell t, return the rook field vector
        """
        M = { (True,True): 0, (True,False): -1, (False,True): 1, (False,False) : 0 }
        return [ M[self.absorbing(t, d, -1),self.absorbing(t, d, 1)] for d in range(0,self.D)]

    def Feedback(self,cycle):
        """
        return {  1   if cycle has an even number of repressors
               { -1   otherwise
        """
        edges = list(zip(cycle[:-1],cycle[1:]))+[(cycle[-1],cycle[0])]
        return functools.reduce(operator.mul, [ 1 if self.network.interaction(i,j) else -1 for (i,j) in edges])

    def CrossingNumber(self, q, cycle):
        edges = list(zip(cycle[:-1],cycle[1:]))+[(cycle[-1],cycle[0])]
        return functools.reduce(operator.add, [ 1 if ((1 if self.network.interaction(i,j) else -1) != q[i]*q[j]) else 0 for (i,j) in edges])

    def RegulationMap(self,s):
        """
        Return a dictionary with key-value pairs
        { (k,v) : s has normal variable k which regulates variable v}
        """
        coords = self.cc.coordinates(s)
        # print(self.cc.boxes())
        # print("RegulationMap")
        # print(s)
        # print(self.closedrightfringe(s))
        # print(coords)
        # print(self.NormalVariables(s))
        # for k in self.NormalVariables(s):
        #     print("normal variable")
        #     print(k)
        #     print(self.network.outputs(k))
        #     print(coords[k])
        #     print(self.network.outputs(k)[coords[k]-1])

        return { k : self.network.outputs(k)[coords[k]-1] for k in self.NormalVariables(s)}

    def isOpaque(self,s):  # not definition in paper, check for bug
        """
        Return True if s is an opaque cubical cell,
          i.e. if RegulationMap(s) is a permutation
        """
        rmap = self.RegulationMap(s)
        return set(rmap.keys())==set(rmap.values())

    def isEquilibriumCell(self,s):
        """
        Return True if s is an equilibrium cell
        """
        if not self.isOpaque(s):
            return False
        for t in self.cc.topstar(s):
            rf = self.RookField(t)
            if any(rf[d] != 0 for d in self.TangentVariables(s)):
                return False
        return True

    def EquilibriumCells(self):
        """
        Returning list of equilibrium cells
        """
        return [ s for s in self.cc if (not self.closedrightfringe(s)) and self.isEquilibriumCell(s) ]

    def CycleDecomposition(self,perm):
        """
        Given a key-value map "perm" defining a permutation,
        return a list of cycles, where each cycle is represented 
        as a list of elements
        e.g.  CycleDecomposition({ 1:1, 3:4, 4:3 }) may return
              [[1],[3,4]]  (or any equivalent answer)
        """
        rmap = dict(perm)
        def ExtractCycle(G):
            cycle = list(G.popitem())
            while cycle[-1] in G: 
                cycle.append(G.pop(cycle[-1]))
            return cycle[:-1]
        cycles = []
        while rmap:
            cycles.append(ExtractCycle(rmap))
        return cycles

    def OpacityCycles(self,s):
        """
        Return list of regulatory cycles for an opaque cubical cell s
        """
        return self.CycleDecomposition(self.RegulationMap(s))


    def Rule1(self):
        """
        Produce all edges, except for edges from real cells into wrap cells
        """
        for edge in self.oc(1):
            (s,t) = self.oc.simplex(edge)
            S = self.dc.dual(s)
            T = self.dc.dual(t)
            if self.blowupfringe(S) or not self.blowupfringe(T):
                self.digraph.add_edge(S,T)
            if self.blowupfringe(T) or not self.blowupfringe(S):
                self.digraph.add_edge(T,S) 

    def Rule2(self):
        """
        Remove edges for which there are transversality theorems from Rule 2
        """
        for edge in self.oc(1):
            #print("Line 87. edge = " + str(edge))
            (s,t) = self.oc.simplex(edge)
            S = self.dc.dual(s)
            T = self.dc.dual(t)
            if self.blowupfringe(S) or self.blowupfringe(T):
                continue
            #print("Line 91. (s,t,S,T) = " + str((s,t,S,T)))
            # s is the index of the lower dimensional cell in the cubical complex self.cc
            # t is the index of the higher dimensional cell in the cubical complex self.cc
            # (s,t) corresponds to an edge in the order complex, which is a wall in the dual-order complex.
            #print("topstar = " + str(self.cc.topstar(t)))
            flowdata = [self.flowdirection(cctopcell,s,t) for cctopcell in self.cc.topstar(t) ]
            #print("Line 96. flowdata = " + str(flowdata))
            if all ( k == 1 for k in flowdata ):
                #print("removing edge " + str((T,S)))
                self.digraph.remove_edge(T, S)
            # else:
            #     print((t,s))
            #     #print("      Edge(" + str(self.cc.barycenter(t)) + ", " + str(self.cc.barycenter(s)) + ")")

            if all ( k == -1 for k in flowdata ):
                #print("removing edge " + str((S,T)))
                self.digraph.remove_edge(S, T)
            # else:
            #     print((s,t))



    def Rule3(self):
        """
        Apply Rule 3 edge removals
        """
        print("** Rule 3 ********************************************")
        for s in self.EquilibriumCells():
            print("=====================")
            print("Checking equilibrium cell s = " + str(s))
            print("  topstar(s) = " + str(self.cc.topstar(s)))
            print("  star(s) = " + str(self.cc.star({s})))
            print("  fringe in star: " + str([(k, str(self.cc.coordinates(k)), self.cc.rightfringe(k)) for k in self.cc.star({s})]))
            print("  dim s = " + str(self.cc.cell_dim(s)))
            print("  Coordinates = " + str(self.cc.coordinates(s)))
            print("  Barycenter = " + str(self.cc.barycenter(s)))
            print("  NormalVariables = " + str(self.NormalVariables(s)))
            print("  TangentVariables = " + str(self.TangentVariables(s)))
            print("  RegulationMap = " + str(self.RegulationMap(s)))
            print("  OpacityCycles = " + str(self.OpacityCycles(s)))
            cycles = self.OpacityCycles(s)
            for t in self.cc.topstar(s):
                print("  Checking top cell t = " + str(t))
                q = self.RelativePositionVector(s,t)
                print("    RelativePositionVector" + str((s,t)) + " = " + str(q))
                unstable = [ (self.CrossingNumber(q,cycle) <= 
                    (len(cycle)+self.Feedback(cycle))/2 + 1) for cycle in cycles]
                for cycle in cycles:
                    edges = list(zip(cycle[:-1],cycle[1:]))+[(cycle[-1],cycle[0])]
                    print("    Cycle = " + str(cycle) )
                    print("      len(cycle) = " + str(len(cycle)))
                    print("      edges = " + str(edges))
                    print("      feedback type = " + str(self.Feedback(cycle)))
                    print("      crossing number = " + str(self.CrossingNumber(q,cycle)))
                    print("      calc = " + str([ (1 if ((1 if self.network.interaction(i,j) else -1) != q[i]*q[j]) else 0,"delta[i,j]=" + str((1 if self.network.interaction(i,j) else -1)),"q[i]=" + str(q[i]),"q[j] = " + str(q[j]),"i = "+str(i),"j = " + str(j)) for (i,j) in edges]))
                print("    Unstable = " + str(unstable))
                S = self.dc.dual(s)
                T = self.dc.dual(t)
                if all(unstable):
                    if T in self.digraph.adjacencies(S):
                        self.digraph.remove_edge(T,S)
                        print("    Removing edge from " + str(t) + " to " + str(s))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      CUT(" + str(self.cc.barycenter(t)) + ", " + str(self.cc.barycenter(s)) + ")")
          
                if not any(unstable):
                    if S in self.digraph.adjacencies(T):
                        self.digraph.remove_edge(S,T)
                        print("    Removing edge from " + str(s) + " to " + str(t))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      CUT(" + str(self.cc.barycenter(s)) + ", " + str(self.cc.barycenter(t)) + ")")

    def Rule4(self):
        """
        Apply Rule 4 edge removals
        """
        print("** Rule 4 ********************************************")
        for s in self.EquilibriumCells():
            cycles = self.OpacityCycles(s)
            for t in self.cc.star({s}):
                q = self.RelativePositionVector(s,t)
                print("    RelativePositionVector" + str((s,t)) + " = " + str(q))
                if any([[q[i] for i in cycle].count(0) > 1 for cycle in cycles]):
                    continue
                qplus = [ x if x != 0 else 1 for x in q ]
                qminus = [ x if x != 0 else -1 for x in q ]           
                unstableplus = [ (self.CrossingNumber(qplus,cycle) <= 
                    (len(cycle)+self.Feedback(cycle))/2 + 1) for cycle in cycles] 
                unstableminus = [ (self.CrossingNumber(qminus,cycle) <= 
                    (len(cycle)+self.Feedback(cycle))/2 + 1) for cycle in cycles]
                unstable = unstableplus + unstableminus
                S = self.dc.dual(s)
                T = self.dc.dual(t)
                if all(unstable):
                    if T in self.digraph.adjacencies(S):
                        self.digraph.remove_edge(T,S)
                if not any(unstable):
                    if S in self.digraph.adjacencies(T):
                        self.digraph.remove_edge(S,T)


    def Rule5(self):
        """
        Apply Rule 5 edge removals
        """
        print("** Rule 5 *#******************************************")
        print(self.EquilibriumCells())
        for s in self.EquilibriumCells():
            print("=====================")
            print("Checking equilibrium cell s = " + str(s))
            print("  topstar(s) = " + str(self.cc.topstar(s)))
            print("  star(s) = " + str(self.cc.star({s})))
            print("  fringe in star: " + str([(k, str(self.cc.coordinates(k)), self.cc.rightfringe(k)) for k in self.cc.star({s})]))
            print("  dim s = " + str(self.cc.cell_dim(s)))
            print("  Coordinates = " + str(self.cc.coordinates(s)))
            print("  Barycenter = " + str(self.cc.barycenter(s)))
            print("  NormalVariables = " + str(self.NormalVariables(s)))
            print("  TangentVariables = " + str(self.TangentVariables(s)))
            print("  RegulationMap = " + str(self.RegulationMap(s)))
            print("  OpacityCycles = " + str(self.OpacityCycles(s)))
            cycles = self.OpacityCycles(s)
            for t in self.cc.star({s}):
                print("  Checking top cell t = " + str(t))
                q = self.RelativePositionVector(s,t)
                print("    RelativePositionVector" + str((s,t)) + " = " + str(q))
                for cycle in cycles:  # debug
                    edges = list(zip(cycle[:-1],cycle[1:]))+[(cycle[-1],cycle[0])]
                    print("    Cycle = " + str(cycle) )
                    print("      len(cycle) = " + str(len(cycle)))
                    print("      edges = " + str(edges))
                    print("      feedback type = " + str(self.Feedback(cycle)))
                    print("      crossing number = " + str(self.CrossingNumber(q,cycle)))
                    print("      calc = " + str([ (1 if ((1 if self.network.interaction(i,j) else -1) != q[i]*q[j]) else 0,"delta[i,j]=" + str((1 if self.network.interaction(i,j) else -1)),"q[i]=" + str(q[i]),"q[j] = " + str(q[j]),"i = "+str(i),"j = " + str(j)) for (i,j) in edges]))
                #print("    Unstable = " + str(unstable)) #debug end

                # skip if not order 2 coface
                if any([[q[i] for i in cycle].count(0) > 1 for cycle in cycles]):
                    continue
                qplus = [ x if x != 0 else 1 for x in q ]
                qminus = [ x if x != 0 else -1 for x in q ]   
                def stability(cycle):
                    unstab1 = (self.CrossingNumber(qplus,cycle) <= 
                        (len(cycle)+self.Feedback(cycle))/2 + 1)
                    unstab2 = (self.CrossingNumber(qminus,cycle) <= 
                        (len(cycle)+self.Feedback(cycle))/2 + 1)
                    print("    Analyzing w_k for cycle " + str(cycle))
                    print("        qplus = " + str(qplus))
                    print("        qminus = " + str(qminus))
                    print("        qplus crossing = " + str(self.CrossingNumber(qplus,cycle)))
                    print("        qminus crossing = " + str(self.CrossingNumber(qminus,cycle)))
                    print("        unstab1 = " + str(unstab1) + "  unstab2 = " + str(unstab2))
                    if unstab1 and unstab2:
                        return 1
                    elif (not unstab1 and not unstab2):
                        return -1
                    else:
                        return 0
                w = [ stability(cycle) for cycle in cycles ]
                print("    w = " + str(w))
                S = self.dc.dual(s)
                T = self.dc.dual(t)
                if all( x <= 0 for x in w) and any(x < 0 for x in w):
                    if S in self.digraph.adjacencies(T):
                        self.digraph.remove_edge(S,T)
                        print("    Removing edge from " + str(s) + " to " + str(t))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      CUT(" + str(self.cc.barycenter(s)) + ", " + str(self.cc.barycenter(t)) + ")")

                if all( x >= 0 for x in w) and any(x > 0 for x in w):
                    if T in self.digraph.adjacencies(S):
                        self.digraph.remove_edge(T,S)
                        print("    Removing edge from " + str(t) + " to " + str(s))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      CUT(" + str(self.cc.barycenter(t)) + ", " + str(self.cc.barycenter(s)) + ")")
          
    def Rule5b(self):
        """
        Apply Rule 5b edge removals
        """
        print("** Rule 5b *#******************************************")
        print(self.EquilibriumCells())
        for s in self.EquilibriumCells():
            print("=====================")
            print("Checking equilibrium cell s = " + str(s))
            print("  topstar(s) = " + str(self.cc.topstar(s)))
            print("  star(s) = " + str(self.cc.star({s})))
            print("  fringe in star: " + str([(k, str(self.cc.coordinates(k)), self.cc.rightfringe(k)) for k in self.cc.star({s})]))
            print("  dim s = " + str(self.cc.cell_dim(s)))
            print("  Coordinates = " + str(self.cc.coordinates(s)))
            print("  Barycenter = " + str(self.cc.barycenter(s)))
            print("  NormalVariables = " + str(self.NormalVariables(s)))
            print("  TangentVariables = " + str(self.TangentVariables(s)))
            print("  RegulationMap = " + str(self.RegulationMap(s)))
            print("  OpacityCycles = " + str(self.OpacityCycles(s)))
            cycles = self.OpacityCycles(s)
            for t in self.cc.star({s}):
                print("  Checking top cell t = " + str(t))
                q = self.RelativePositionVector(s,t)
                print("    RelativePositionVector" + str((s,t)) + " = " + str(q))
                for cycle in cycles:  # debug
                    edges = list(zip(cycle[:-1],cycle[1:]))+[(cycle[-1],cycle[0])]
                    print("    Cycle = " + str(cycle) )
                    print("      len(cycle) = " + str(len(cycle)))
                    print("      edges = " + str(edges))
                    print("      feedback type = " + str(self.Feedback(cycle)))
                    print("      crossing number = " + str(self.CrossingNumber(q,cycle)))
                    print("      calc = " + str([ (1 if ((1 if self.network.interaction(i,j) else -1) != q[i]*q[j]) else 0,"delta[i,j]=" + str((1 if self.network.interaction(i,j) else -1)),"q[i]=" + str(q[i]),"q[j] = " + str(q[j]),"i = "+str(i),"j = " + str(j)) for (i,j) in edges]))
                #print("    Unstable = " + str(unstable)) #debug end

                # skip if not order 2 coface
                if any([[q[i] for i in cycle].count(0) > 1 for cycle in cycles]):
                    continue
                qplus = [ x if x != 0 else 1 for x in q ]
                qminus = [ x if x != 0 else -1 for x in q ]   
                def stability(cycle):
                    unstab1 = (self.CrossingNumber(qplus,cycle) <= 
                        (len(cycle)+self.Feedback(cycle))/2 + 1)
                    unstab2 = (self.CrossingNumber(qminus,cycle) <= 
                        (len(cycle)+self.Feedback(cycle))/2 + 1)
                    print("    Analyzing w_k for cycle " + str(cycle))
                    print("        qplus = " + str(qplus))
                    print("        qminus = " + str(qminus))
                    print("        qplus crossing = " + str(self.CrossingNumber(qplus,cycle)))
                    print("        qminus crossing = " + str(self.CrossingNumber(qminus,cycle)))
                    print("        unstab1 = " + str(unstab1) + "  unstab2 = " + str(unstab2))
                    if unstab1 and unstab2:
                        return 1
                    elif (not unstab1 and not unstab2):
                        return -1
                    else:
                        return 0
                w = [ stability(cycle) for cycle in cycles ]
                print("    w = " + str(w))
                S = self.dc.dual(s)
                T = self.dc.dual(t)
                if all( x <= 0 for x in w) and any(x < 0 for x in w):
                    if S in self.digraph.adjacencies(T):
                        self.digraph.remove_edge(S,T)
                        print("    Removing edge from " + str(s) + " to " + str(t))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      CUT(" + str(self.cc.barycenter(s)) + ", " + str(self.cc.barycenter(t)) + ")")

                if all( x >= 0 for x in w) and any(x > 0 for x in w):
                    if T in self.digraph.adjacencies(S):
                        self.digraph.remove_edge(T,S)
                        print("    Removing edge from " + str(t) + " to " + str(s))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      CUT(" + str(self.cc.barycenter(t)) + ", " + str(self.cc.barycenter(s)) + ")")
          
    def Rule6(self):
        """
        Apply Rule 6 edge removals
        """
        print("** Rule 6 ********************************************")
        print(self.EquilibriumCells())
        for s in self.EquilibriumCells():
            print("=====================")
            print("Checking equilibrium cell s = " + str(s))
            print("  topstar(s) = " + str(self.cc.topstar(s)))
            print("  star(s) = " + str(self.cc.star({s})))
            print("  fringe in star: " + str([(k, str(self.cc.coordinates(k)), self.cc.rightfringe(k)) for k in self.cc.star({s})]))
            print("  dim s = " + str(self.cc.cell_dim(s)))
            print("  Coordinates = " + str(self.cc.coordinates(s)))
            print("  Barycenter = " + str(self.cc.barycenter(s)))
            print("  NormalVariables = " + str(self.NormalVariables(s)))
            print("  TangentVariables = " + str(self.TangentVariables(s)))
            print("  RegulationMap = " + str(self.RegulationMap(s)))
            print("  OpacityCycles = " + str(self.OpacityCycles(s)))
            cycles = self.OpacityCycles(s)
            for t in self.cc.star({s}):
                print("  Checking top cell t = " + str(t))
                q = self.RelativePositionVector(s,t)
                print("    RelativePositionVector" + str((s,t)) + " = " + str(q))
                for cycle in cycles:  # debug
                    edges = list(zip(cycle[:-1],cycle[1:]))+[(cycle[-1],cycle[0])]
                    print("    Cycle = " + str(cycle) )
                    print("      len(cycle) = " + str(len(cycle)))
                    print("      edges = " + str(edges))
                    print("      feedback type = " + str(self.Feedback(cycle)))
                    print("      crossing number = " + str(self.CrossingNumber(q,cycle)))
                    print("      calc = " + str([ (1 if ((1 if self.network.interaction(i,j) else -1) != q[i]*q[j]) else 0,"delta[i,j]=" + str((1 if self.network.interaction(i,j) else -1)),"q[i]=" + str(q[i]),"q[j] = " + str(q[j]),"i = "+str(i),"j = " + str(j)) for (i,j) in edges]))
                #print("    Unstable = " + str(unstable)) #debug end

                # # skip if not order 2 coface
                # if any([[q[i] for i in cycle].count(0) > 1 for cycle in cycles]):
                #     continue
                qplus = [ x if x != 0 else 1 for x in q ]
                qminus = [ x if x != 0 else -1 for x in q ]   
                def stability(cycle):
                    unstab1 = (self.CrossingNumber(qplus,cycle) <= 
                        (len(cycle)+self.Feedback(cycle))/2 + 1)
                    unstab2 = (self.CrossingNumber(qminus,cycle) <= 
                        (len(cycle)+self.Feedback(cycle))/2 + 1)
                    print("    Analyzing w_k for cycle " + str(cycle))
                    print("        qplus = " + str(qplus))
                    print("        qminus = " + str(qminus))
                    print("        qplus crossing = " + str(self.CrossingNumber(qplus,cycle)))
                    print("        qminus crossing = " + str(self.CrossingNumber(qminus,cycle)))
                    print("        unstab1 = " + str(unstab1) + "  unstab2 = " + str(unstab2))
                    if unstab1 and unstab2:
                        return 1
                    elif (not unstab1 and not unstab2):
                        return -1
                    else:
                        return 0
                w = [ stability(cycle) for cycle in cycles ]
                print("    w = " + str(w))
                S = self.dc.dual(s)
                T = self.dc.dual(t)
                if all( x <= 0 for x in w) and any(x < 0 for x in w):
                    if S in self.digraph.adjacencies(T):
                        self.digraph.remove_edge(S,T)
                        print("    Removing edge from " + str(s) + " to " + str(t))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      CUT(" + str(self.cc.barycenter(s)) + ", " + str(self.cc.barycenter(t)) + ")")

                if all( x >= 0 for x in w): # and any(x > 0 for x in w):
                    if T in self.digraph.adjacencies(S):
                        self.digraph.remove_edge(T,S)
                        print("    Removing edge from " + str(t) + " to " + str(s))
                        print("      barycenter(" + str(t) + ") = " + str(self.cc.barycenter(t)))
                        print("      barycenter(" + str(s) + ") = " + str(self.cc.barycenter(s)))
                        print("      CUT(" + str(self.cc.barycenter(t)) + ", " + str(self.cc.barycenter(s)) + ")")
          

    def __init__(self, p, level = 0):
        # initialize complex, graph, supporting data structures
        self.p = p
        self.network = self.p.network()
        self.D = self.network.size()
        self.cc = CubicalComplex([ x + 1 for x in self.network.domains()])
        self.digraph = DirectedAcyclicGraph()
        self.labelling = p.labelling()
        self.limits = self.network.domains()
        self.pv = [1]
        for i in self.limits:
            self.pv.append(self.pv[-1]*i)
        self.oc = OrderComplex(self.cc)
        self.dc = DualComplex(self.oc)
        for cell in self.dc(self.D):
            self.digraph.add_vertex(cell)

        # Apply rules
        self.Rule1()
        self.Rule2()
        if level == 3:
            self.Rule3()
        if level == 4:
            self.Rule3()
            self.Rule4()
        if level == 5:
            self.Rule3()
            self.Rule5()
        if level == "5b":
            self.Rule3()
            self.Rule5b()
        if level == 6:
            self.Rule3()
            self.Rule6()



# SaveDatabaseJSON.py
# Marcio Gameiro
# MIT LICENSE
# 2021-07-11

import DSGRN
import pychomp2
import itertools
import json

__all__ = ['DatabaseJSON', 'SaveDatabaseJSON']

def dsgrn_cell_to_cc_cell_map(network):
    """Return a mapping from the top dimensional cells
    in the DSGRN complex to the top dimensional
    cells in the pychomp2 cubical complex.
    """
    # Construct a cubical complex using pychomp2. A cubical complex in pychomp2
    # does not contain the rightmost boundary, so make one extra layer of
    # cubes and ignore the last layer (called rightfringe in pychomp2).
    cubical_complex = pychomp2.CubicalComplex([n + 1 for n in network.domains()])
    dimension = network.size()
    # Mapping from DSGRN top cells to cc top cells
    cell2cc_cell = {}
    # DSGRN only uses the top dimensional cells to
    # construct the state transiton graph and the
    # corresponding morse sets and these cells are
    # indexed from 0 to n-1, where n is the number
    # of top dimensional cells. Hence we can get a
    # mapping by counting the cells in the cc.
    dsgrn_index = 0
    for cell_index in cubical_complex(dimension):
        # Ignore fringe cells
        if cubical_complex.rightfringe(cell_index):
            continue
        cell2cc_cell[dsgrn_index] = cell_index
        dsgrn_index += 1
    return cell2cc_cell

def network_json(network):
    """Return json data for network."""
    D = network.size()
    nodes = [] # Get network nodes
    for d in range(D):
        node = {"id" : network.name(d)}
        nodes.append(node)
    # Get network edges
    edges = [(u, v) for u in range(D) for v in network.outputs(u)]
    links = []
    for (u, v) in edges:
        edge_type = 1 if network.interaction(u, v) else -1
        link = {"source" : network.name(u),
                "target" : network.name(v),
                "type": edge_type}
        links.append(link)
    # Get number of out-edges (treat the no out edge case as one out edge)
    num_edges = sum([len(network.outputs(d)) if network.outputs(d) else 1 for d in range(D)])
    param_dim = 3 * num_edges + D
    network_json_data = {"network" : {"nodes" : nodes, "links" : links, "parameter_dim" : param_dim}}
    return network_json_data

def parameter_graph_json(parameter_graph, vertices=None, verts_colors=None, thres_type=''):
    """Return json data for parameter graph."""
    # Get list of vertices if none
    if vertices == None:
        vertices = list(range(parameter_graph.size()))
    # Set empty dictionary for verts_colors if none
    if verts_colors == None:
        verts_colors = {}
    # Set thres_type to '' if not 'T'
    if thres_type != 'T':
        thres_type = '' # Uses the default 't' type
    all_edges = [(u, v) for u in vertices for v in parameter_graph.adjacencies(u, 'codim1') if v in vertices]
    # Remove double edges (all edges are double)
    edges = [(u, v) for (u, v) in all_edges if u > v]
    nodes = []
    for v in vertices:
        v_color = verts_colors[v] if v in verts_colors else ""
        v_ineqs = parameter_graph.parameter(v).partialorders(thres_type).split('\n')
        node = {"id" : v, "color" : v_color, "inequalities" : v_ineqs}
        # node = {"id" : str(v), "color" : v_color, "inequalities" : v_ineqs}
        nodes.append(node)
    links = []
    for (u, v) in edges:
        link = {"source" : u, "target" : v}
        # link = {"source" : str(u), "target" : str(v)}
        links.append(link)
    parameter_graph_json_data = {"parameter_graph" : {"nodes" : nodes, "links" : links}}
    return parameter_graph_json_data

def cubical_complex_json(network):
    """Return json data for cubical complex."""
    # Get complex dimension
    dimension = network.size()
    # Construct a cubical complex using pychomp2. A cubical complex in pychomp2
    # does not contain the rightmost boundary, so make one extra layer of
    # cubes and ignore the last layer (called rightfringe in pychomp2).
    cubical_complex = pychomp2.CubicalComplex([n + 1 for n in network.domains()])
    # Get vertices coordinates and set a
    # mapping from coords to its index in
    # the list of coordinates.
    verts_coords = []
    coords2idx = {}
    # Get the coords of all cells of dimension 0.
    # The 0-dim cells in a cubical complex in pychomp2
    # are indexed from 0 to n-1, where n is the number
    # of 0-dim cells. Hence the cell_index coincides
    # with the index of coords in the list verts_coords.
    for cell_index in cubical_complex(0):
        coords = cubical_complex.coordinates(cell_index)
        coords2idx[tuple(coords)] = cell_index
        verts_coords.append(coords)
    cells = [] # Get the cell complex data
    for cell_index in cubical_complex:
        # Ignore fringe cells
        if cubical_complex.rightfringe(cell_index):
            continue
        # Get this cell dimension
        cell_dim = cubical_complex.cell_dim(cell_index)
        # Get coords of the lower corner of the box
        coords_lower = cubical_complex.coordinates(cell_index)
        # Get index of vertex corresponding to these coords
        # Due to the way pychomp2 index the 0-dim cells we get
        # that idx_lower == cell_index (see coords2idx above).
        idx_lower = coords2idx[tuple(coords_lower)]
        # Get the shape of this cell (see pychomp2)
        shape = cubical_complex.cell_shape(cell_index)
        # Add 1 to the appropriate entries to get coords of the upper corner
        coords_upper = [coords_lower[d] + (1 if shape & (1 << d) != 0 else 0) for d in range(dimension)]
        # Get index of vertex corresponding to these coords
        idx_upper = coords2idx[tuple(coords_upper)]
        # Get indices of coords that have extent
        ind_extent = [d for d in range(dimension) if shape & (1 << d) != 0]
        if cell_dim == 0:
            cell_verts = [idx_lower]
        elif cell_dim == 1:
            cell_verts = [idx_lower, idx_upper]
        elif cell_dim == 2:
            # Index of vertex 1
            idx1 = idx_lower
            # Coords and index of vertex 2
            coords_v2 = [coord for coord in coords_lower]
            coords_v2[ind_extent[0]] += 1
            idx2 = coords2idx[tuple(coords_v2)]
            # Index of vertex 3
            idx3 = idx_upper
            # Coords and index of vertex 4
            coords_v4 = [coord for coord in coords_lower]
            coords_v4[ind_extent[1]] += 1
            idx4 = coords2idx[tuple(coords_v4)]
            cell_verts = [idx1, idx2, idx3, idx4]
        elif cell_dim == 3: # cell_dim == dimension == 3
            # First get vertices of unit cube as cartesian product of {0, 1}
            u_verts = list(itertools.product((0, 1), repeat=cell_dim))
            # Add verts of unit cube to coords_lower to get verts of this cell
            cell_verts = []
            for u in u_verts:
                coords_vert = [sum(x) for x in zip(u, coords_lower)]
                idx_vert = coords2idx[tuple(coords_vert)]
                cell_verts.append(idx_vert)
        else: # Ignore cell if dim > 3
            continue
        cell = {"cell_dim" : cell_dim, "cell_index" : cell_index, "cell_verts" : cell_verts}
        cells.append(cell)
    complex_json_data = {"complex" : {"dimension" : dimension,
                                      "verts_coords" : verts_coords,
                                      "cells" : cells} }
    return complex_json_data

def morse_graph_json(morse_graph):
    """Return json data for Morse graph."""

    def vertex_rank(u):
        """Return how many levels down of children u have."""
        children = [v for v in morse_graph.poset().children(u)]
        if len(children) == 0:
            return 0
        return 1 + max([vertex_rank(v) for v in children])

    # Get list of Morse nodes
    morse_nodes = range(morse_graph.poset().size())
    morse_graph_data = [] # Morse graph data
    for morse_node in morse_nodes:
        label = morse_graph.annotation(morse_node)[0]
        adjacencies = morse_graph.poset().children(morse_node)
        morse_node_data = {"node" : morse_node,
                           "rank" : vertex_rank(morse_node),
                           "label" : label,
                           "adjacencies" : adjacencies}
        morse_graph_data.append(morse_node_data)
    morse_graph_json_data = {"morse_graph" : morse_graph_data}
    return morse_graph_json_data

def morse_sets_json(network, morse_graph, morse_decomposition):
    """Return json data for Morse sets."""
    # Get a mapping from DSGRN top cells to cc top cells
    cell2cc_cell = dsgrn_cell_to_cc_cell_map(network)
    # Get list of Morse nodes
    morse_nodes = range(morse_decomposition.poset().size())
    # Permutation that gives node index from Morse set index
    permutation = morse_graph.permutation()
    morse_sets_data = [] # Morse sets data
    for morse_node in morse_nodes:
        morse_cells = [cell2cc_cell[c] for c in morse_decomposition.morseset(morse_node)]
        # Get Morse graph node index
        morse_graph_node = permutation[morse_node]
        morse_set = {"index" : morse_graph_node, "cells" : morse_cells}
        morse_sets_data.append(morse_set)
    morse_sets_json_data = {"morse_sets" : morse_sets_data}
    return morse_sets_json_data

def equilibrium_cells_json(parameter, morse_sets_data):
    """Return json data for Equilibrium cells."""
    # Get indices of all equilibrium cells
    eq_cells = DSGRN.EquilibriumCells(parameter, 'all', 'index')
    # Get Morse cells from the Morse sets
    morse_cells = [cell for morse_set in morse_sets_data for cell in morse_set["cells"]]
    # Get equilibrium cells that are not part of a Morse set
    equilibrium_cells = [cell for cell in eq_cells if cell not in morse_cells]
    eq_cells_json_data = {"equilibrium_cells" : equilibrium_cells}
    return eq_cells_json_data

def state_transition_graph_json(network, domain_graph):
    """Return json data for state transiton graph."""
    # Get a mapping from DSGRN top cells to cc top cells
    cell2cc_cell = dsgrn_cell_to_cc_cell_map(network)
    # Get state transition graph vertices
    stg_vertices = range(domain_graph.digraph().size())
    stg = [] # State transition graph
    for v in stg_vertices:
        adjacencies = [cell2cc_cell[u] for u in domain_graph.digraph().adjacencies(v)]
        node_adjacency_data = {"node" : cell2cc_cell[v], "adjacencies" : adjacencies}
        stg.append(node_adjacency_data)
    stg_json_data = {"stg" : stg}
    return stg_json_data

def DatabaseJSON(network, param_indices=None, verts_colors=None, eq_cells=False, thres_type=''):
    if network.size() not in [2, 3]:
        print('Only available for dimensions 2 and 3!')
        return
    parameter_graph = DSGRN.ParameterGraph(network)
    # Use all parameter indices if None
    if param_indices == None:
        param_indices = range(parameter_graph.size())
    network_json_data = network_json(network)
    cell_complex_json_data = cubical_complex_json(network)
    param_graph_json_data = parameter_graph_json(parameter_graph, param_indices, verts_colors, thres_type)
    dynamics_database = [] # Dynamics database
    for par_index in param_indices:
        # Compute DSGRN dynamics
        parameter = parameter_graph.parameter(par_index)
        domain_graph = DSGRN.DomainGraph(parameter)
        morse_decomposition = DSGRN.MorseDecomposition(domain_graph.digraph())
        morse_graph = DSGRN.MorseGraph(domain_graph, morse_decomposition)
        morse_graph_json_data = morse_graph_json(morse_graph)
        morse_sets_json_data = morse_sets_json(network, morse_graph, morse_decomposition)
        stg_json_data = state_transition_graph_json(network, domain_graph)
        if eq_cells: # Include equilibrium cells if true
            eq_cells_json_data = equilibrium_cells_json(parameter, morse_sets_json_data["morse_sets"])
        else:
            eq_cells_json_data = {"equilibrium_cells" : []}
        # Dynamics data for this parameter
        dynamics_json_data = {"parameter" : par_index,
                              "morse_graph" : morse_graph_json_data["morse_graph"],
                              "morse_sets" : morse_sets_json_data["morse_sets"],
                              "equilibrium_cells" : eq_cells_json_data["equilibrium_cells"],
                              "stg" : stg_json_data["stg"]}
        dynamics_database.append(dynamics_json_data)
    morse_graph_database = {"network" : network_json_data["network"],
                            "complex" : cell_complex_json_data["complex"],
                            "parameter_graph" : param_graph_json_data["parameter_graph"],
                            "dynamics_database" : dynamics_database}
    return morse_graph_database

def SaveDatabaseJSON(network, database_fname, param_indices=None, verts_colors=None,
                     eq_cells=False, thres_type=''):
    morse_graph_database = DatabaseJSON(network, param_indices, verts_colors, eq_cells, thres_type)
    # Save database to a file
    with open(database_fname, 'w') as outfile:
        json.dump(morse_graph_database, outfile)



# Graphics.py
# Shaun Harker
# 2018-03-27
# MIT LICENSE

import IPython.display

def makehtml(item):
    if type(item) == type([]) or type(item) == type(()):
        return '<div>[' + ','.join([makehtml(i) for i in item]) + ']</div>'
    try:
        itemhtml = item._repr_html_()
    except AttributeError:
        try:
            itemhtml = item._repr_svg_()
        except AttributeError:
            itemhtml = str(item)
    return itemhtml  

class View:
    def __init__(self, obj):
        self.html = makehtml(obj)
    def _repr_html_(self):
        return self.html

class Table:
    def __init__(self, headers, rows, etc = None):
        self.headers = headers
        self.rows = rows
        self.etc = etc
    def tablestring(self, rows):
        html = '<div style="border:1px black">'
        html += '<table>'
        html += "<tr>"
        for item in self.headers:
            html += '<th>' 
            html += '<div style="text-align:center;">'
            html += makehtml(item) + '</div></th>'
        html += "</tr>"
        for row in rows:
            html += "<tr>"
            for item in row:
                html += '<td>'
                html += '<div style="text-align:center;">'
                html += makehtml(item) + '</div></td>'
            html += "</tr>"
        html += '</div></table></div>'
        return html
    def _repr_html_(self):
        return self.tablestring(self.rows)
    def view(self, rowindices = None):
        if not rowindices:
            rowindices = range(0,len(self.rows))
        if type(rowindices) == type(0): # single row table
            rowindices = [rowindices]
        return View(self.tablestring((self.rows[i] for i in rowindices)))



# FixedPointTables.py
# MIT LICENSE 2016
# Shaun Harker

# This file contains routines used by 
#   SingleFixedPointQuery.py  and
#   DoubleFixedPointQuery.py
# in order to create SQL tables to support their queries.

from DSGRN.Query.Logging import LogToSTDOUT

def FPString(i, j, database):
  terms = [ "_" for k in range(0, database.D) ]
  terms[i] = str(j)  
  expression = "Label like 'FP { " + ', '.join(terms) + "%'";
  return expression

def buildQueryExpression(bounds, database):
  expressions = []
  for i in range(0,database.D):
    networknodename = database.names[i]
    lowerbound = 0
    upperbound = len(database.network.outputs(i))
    if networknodename in bounds:
      varbounds = bounds[networknodename]
      if type(varbounds) == int :
        lowerbound = varbounds
        upperbound = varbounds
      else:
        lowerbound = varbounds[0]
        upperbound = varbounds[1]
    expression = " or ".join([ FPString(i,j,database) for j in range(lowerbound, upperbound+1)])
    expressions.append(expression)
  return expressions

def MatchQuery(bounds, outputtablename, database):
  # Parse the command line
  LogToSTDOUT("MatchQuery(" + str(bounds) + ", " + str(outputtablename) + ")")
  c = database.conn.cursor()
  expressions = buildQueryExpression(bounds, database)
  LogToSTDOUT("MatchQuery :: built expressions " + str(expressions) )
  N = len(expressions)
  for i, expression in enumerate(expressions):
    oldtable = 'tmpMatches' + str(i)
    if i == 0: oldtable = "MorseGraphAnnotations"
    newtable = 'tmpMatches' + str(i+1)
    if i == N-1: newtable = outputtablename
    sql_string = 'create temp table ' + newtable + ' as select * from ' + oldtable + ' where ' + expression + ';'
    LogToSTDOUT("MatchQuery :: " + sql_string)
    c.execute(sql_string)
    if i > 0: c.execute('drop table ' + oldtable);
  LogToSTDOUT("MatchQuery :: constructed")






# PosetOfExtrema.py
# MIT LICENSE
# Shaun Harker 2016-12-27
#
# Marcio Gameiro
# 2021-08-04

from DSGRN._dsgrn import *

class PosetOfExtrema(Pattern):
  def __init__(self, network, events, orderings):
    """
    Initializes a poset of extrema
      network is a DSGRN network object
      events is a list of tuples of the form (NAME,'min') or (NAME,'max') for NAME a
                name of a node in the network
      orderings is list of event orderings (the code will perform transitive closure
                automatically), i.e. a list of pairs (i,j) such that events[i] occurs
                before events[j]
    """
    # Save variables
    self.network_ = network
    self.labels_ = [event[0] + " " + event[1] for event in events]
    # Construct Digraph from event and ordering data
    digraph = Digraph()
    for event in events:
      digraph.add_vertex()
    for edge in orderings:
      digraph.add_edge(edge[0], edge[1])
    self.poset_ = Poset(digraph)
    # Find final occurrence of each variable in poset
    latest_event = {}
    label = {}
    for i, event in enumerate(events):
      variable = network.index(event[0])
      if (variable not in latest_event) or self.poset_.compare(latest_event[variable], i):
        latest_event[variable] = i
        label[variable] = 2 ** (variable + network.size()) if event[1].lower() == "min" else 2 ** variable
    final_label = 0
    for variable in range(network.size()):
      if variable not in latest_event:
        label[variable] = 0
      final_label += label[variable]
    events_for_pattern = [network.index(event[0]) for event in events]
    # Create "Pattern" object instance
    super(PosetOfExtrema, self).__init__(self.poset_, events_for_pattern, final_label, self.network_.size())

  def graphviz(self):
    """
    Return graphviz string for visualization
    """ 
    N = len(self.labels_)
    gv_vertices = [ '"' + str(i) + '" [label="' + self.labels_[i] + '"];' for i in range(N) ]
    gv_edges = [ '"' + str(u) + '" -> "' + str(v) + '";' for u in range(N) for v in self.poset_.children(u) ]
    return 'digraph {\n' + '\n'.join(gv_vertices) + '\n' + '\n'.join(gv_edges) + '\n}\n'



### __init__.py
### MIT LICENSE 2018 Shaun Harker
### MIT LICENSE 2021 Marcio Gameiro

from DSGRN._dsgrn import *
from DSGRN.SubdomainGraph import *
from DSGRN.BlowupGraph import *
from DSGRN.Graphics import *
from DSGRN.Query.Graph import *
from DSGRN.Query.Database import *
from DSGRN.Query.Hexcodes import *
from DSGRN.Query.MonostableQuery import *
from DSGRN.Query.BistableQuery import *
from DSGRN.Query.MultistableQuery import *
from DSGRN.Query.NstableQuery import *
from DSGRN.Query.SingleFixedPointQuery import *
from DSGRN.Query.DoubleFixedPointQuery import *
from DSGRN.Query.MonostableFixedPointQuery import *
from DSGRN.Query.SingleGeneQuery import *
from DSGRN.Query.InducibilityQuery import *
from DSGRN.Query.HysteresisQuery import *
from DSGRN.Query.PhenotypeQuery import *
from DSGRN.Query.PosetOfExtrema import *
from DSGRN.Query.Logging import *
from DSGRN.Query.StableFCQuery import *
from DSGRN.Query.ComputeSingleGeneQuery import *
from DSGRN.EssentialParameterNeighbors import *
from DSGRN.ParameterPartialOrders import *
from DSGRN.ParameterFromSample import *
from DSGRN.SaveDatabaseJSON import *
from DSGRN.EquilibriumCells import *
import sys
import os
import pickle

configuration().set_path(os.path.dirname(__file__) + '/Resources')



import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # Required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cfg = 'Debug' if self.debug else 'Release'

        # Get generator info (CMake lets you override generator)
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={}".format(extdir),
            "-DPYTHON_EXECUTABLE={}".format(sys.executable),
            "-DEXAMPLE_VERSION_INFO={}".format(self.distribution.get_version()),
            "-DCMAKE_BUILD_TYPE={}".format(cfg),  # Not used on MSVC
            "-DUSER_INCLUDE_PATH=./src/DSGRN/_dsgrn/include"
        ]

        build_args = ['--config', cfg]

        if self.compiler.compiler_type == "msvc":
            # Single config generators are handled normally
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec in the generator name
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
                ]
                build_args += ["--config", cfg]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())

        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)

# Read the contents of README.md file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='DSGRN',
    version='1.4.1',
    author='Marcio Gameiro',
    author_email='marciogameiro@gmail.com',
    description='DSGRN (Dynamic Signatures Generated by Regulatory Networks)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_dir = {'': 'src'},
    ext_package='DSGRN',
    ext_modules=[CMakeExtension('_dsgrn')],
    packages=['DSGRN'],
    entry_points={'console_scripts': ['Signatures=DSGRN.Signatures:main [MPI]']},
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    url = 'https://github.com/marciogameiro/DSGRN',
    include_package_data = True,
    install_requires=['scipy', 'matplotlib', 'numpy', 'graphviz', 'progressbar2', 'jupyter', 'pychomp2'],
    extras_require={
        'MPI':  ["mpi4py"]
    }
)



# SingleFixedPointQuery.py
# MIT LICENSE 2016
# Shaun Harker

from DSGRN.Query.FixedPointTables import *
from DSGRN.Query.Logging import LogToSTDOUT

class SingleFixedPointQuery:
  def __init__ (self, database, bounds):
    self.database = database
    LogToSTDOUT("SingleFixedPointQuery :: initializing")
    LogToSTDOUT("SingleFixedPointQuery :: calling MatchQuery")
    MatchQuery(bounds,"Matches",database)
    # Final query and print results
    c = database.conn.cursor()
    LogToSTDOUT("SingleFixedPointQuery :: select MorseGraphIndex from Matches;")
    self.set_of_matches = set([ row[0] for row in c.execute('select MorseGraphIndex from Matches;')])
    #result = [row[0] for row in c.execute('select distinct ParameterIndex from Matches natural join Signatures order by ParameterIndex;')]
    LogToSTDOUT("SingleFixedPointQuery :: drop table Matches;")
    c.execute('drop table Matches');
    LogToSTDOUT("SingleFixedPointQuery :: constructed")


  def matches(self):
    """
    Return entire set of matches
    """
    return self.set_of_matches

  def __call__ (self, morsegraphindex ):
    """ 
    Test if a single mgi is in the set of matches
    """
    return morsegraphindex in self.set_of_matches




# NstableQuery.py
# Bree Cummins
# Edited from
# MultistableQuery.py
# MIT LICENSE 2016
# Shaun Harker

class NstableQuery:
  """
  NstableQuery
    Query to check if morse graph index corresponds to a Morse graph with at least N minimal fixed points
  """
  def __init__ (self, database, N ):
    self.database = database
    c = database.conn.cursor()
    sqlexpression = "select MorseGraphIndex from (select MorseGraphIndex, count(*) as StableCount from (select MorseGraphIndex,Vertex from MorseGraphAnnotations where Label like 'FP%' except select MorseGraphIndex,Source from MorseGraphEdges) group by MorseGraphIndex) where StableCount>{};".format(N-1)
    database.NQuery = frozenset([ row[0] for row in c.execute(sqlexpression) ])

  def matches(self):
    """
    Return entire set of matches
    """
    return self.database.NQuery

  def __call__ (self, morsegraphindex ):
    """ 
    Test if a single mgi is in the set of matches
    """
    return morsegraphindex in self.database.NQuery



import graphviz

from DSGRN._dsgrn import *

def DrawGraph(g):
  """
  Return an object which renders in Notebook visualizations.
  Works with any input g with a "graphviz" method returning a valid
  graphviz string
  """
  return graphviz.Source(g.graphviz())

def DrawGraphWithHighlightedPath(g, path):
  """
  Return an object which renders in Notebook visualizations.
  Works for "SearchGraph" "PatternGraph" and "MatchingGraph".
  If given a python integer list it can convert for "SearchGraph" and "PatternGraph"
  Currently "MatchingGraph" requires an IntPairList input
  """
  return graphviz.Source(g.graphviz_with_highlighted_path(path))

class Graph:
  def __init__(self, vertices, edges):
    """
    Construct graph from set of vertices and collection of edges
    Note: vertices and edges can be presented as either lists of sets;
          they will be interally converted to sets.
    """
    self.vertices = set(vertices)
    self.edges = set(edges) 
    self.adjacency_lists = {}
    for s,t in self.edges:
      if not s in self.vertices:
        self.vertices.add(s)
      if not t in self.vertices:
        self.vertices.add(t)
      if not s in self.adjacency_lists:
          self.adjacency_lists[s] = []
      self.adjacency_lists[s].append(t) 
    self.vertexname = {}
    for i, v in enumerate(self.vertices):
      self.vertexname[v] = 'X' + str(i)

  def clone(self):
    """
    Return a deep copy
    """
    return Graph(self.vertices, self.edges)

  def _repr_svg_(self):
    """
    Return svg representation suitable for Notebook visualization
    """
    return graphviz.Source(self.graphviz())._repr_svg_()

  def graphviz(self):
    """
    Return graphviz string for graph
    """
    return 'digraph {' + \
  '\n'.join([ '"' + self.vertexname[v] + '" [label="' + self.label(v) + '";style="filled";fillcolor="' + self.color(v) + '"];' for v in self.vertices ]) + \
   '\n' + '\n'.join([ '"' + self.vertexname[u]  + '" -> "' + self.vertexname[v]  + '";' for (u, v) in self.edges ]) + \
   '\n' + '}\n'

  def adjacencies(self, p):
    """
    Return adjacency list for vertex p (i.e. { q : (p,q) \in edges })
    """
    return self.adjacency_lists.get(p,[])

  def reachable(self, p, q):
    """
    Return "true" if there is a path from p to q
    """
    explored = set()
    items = [p]
    while items:
      x = items.pop()
      if x == q: return True
      if x in explored: continue
      explored.add(x)
      items.extend([ y for y in self.adjacencies(x) if y not in explored ])
    return False

  def predicate_reachable(self, p, allowed, terminal):
    """
    Return "true" if there is a path from the vertex p to some vertex q such that
    each node in the path (except possibly q) satisfies allowed(p) and q satisfies terminal(q)
    """
    if not allowed(p):
      return False
    explored = set()
    items = [p]
    while items:
      x = items.pop()
      if terminal(x): return True
      if x in explored: continue
      if not allowed(x): continue
      explored.add(x)
      items.extend([ y for y in self.adjacencies(x) if y not in explored ])
    return False

  def numberOfPaths(self, p, q):
    """
    Compute the number of paths from "p" to "q" in the graph described by "adjacency_lists"
    TODO: this appears to be unexpectedly slow
    """
    cache = {q : 1}
    subresult = lambda x : cache.get(x, sum([subresult(y) for y in self.adjacencies(x)]))
    return subresult(p)

  def subgraph(self, predicate):
    """
    Extract induced subgraph consisting of vertices satisfying "predicate"
    """
    return Graph({ v for v in self.vertices if predicate(v)}, { (s,t) for (s,t) in self.edges if predicate(s) and predicate(t)})

  def unavoidable(self, p, q, predicate):
    """
    Return "true" if every path in a graph from "p" to "q" must satisfy "predicate" at one or more vertices.
    """
    return self.subgraph(lambda v : not predicate(v)).numberOfPaths(p,q) == 0

  def label(self, v):
    """
    Return a label string to be used when displaying graph
    """
    return str(v)

  def color(self, v):
    """
    Return a fillcolor to be used when displaying graph
    """
    return "blue"



# Hexcodes.py
# MIT LICENSE 2016
# Shaun Harker 

def isPowerOfTwo (n):
  return (n & (n-1) == 0) and (n != 0)

def isAdjacentHexcode(hexcode1, hexcode2):
  return int(hexcode1, 16) < int(hexcode2, 16) and isPowerOfTwo( int(hexcode1, 16) ^ int(hexcode2, 16) )

def essential(hexcode,n,m):
  i = int(hexcode, 16)
  binning = [sum ([ 1 if i & 2**k else 0 for k in range(j*m,(j+1)*m)]) for j in range(0,2**n)]
  if not 0 in binning:
    return False 
  if not m in binning:
    return False
  N = 2**n 
  for d in range(0,n):
    bit = 2**d 
    flag = False
    for x in range(0, N):
      if binning[x] != binning[x ^ bit]:
        flag = True
        break
    if not flag:
      return False
  return True
